package eu.mhsl.craftattack.spawn.appliances.countdown;

import eu.mhsl.craftattack.spawn.Main;
import eu.mhsl.craftattack.spawn.appliance.Appliance;
import eu.mhsl.craftattack.spawn.appliance.ApplianceCommand;
import eu.mhsl.craftattack.spawn.appliances.countdown.command.ProjectStartCancelCommand;
import eu.mhsl.craftattack.spawn.appliances.countdown.command.ProjectStartCommand;
import eu.mhsl.craftattack.spawn.appliances.countdown.command.ProjectStartResetCommand;
import eu.mhsl.craftattack.spawn.appliances.countdown.listener.PlayerInvincibleListener;
import eu.mhsl.craftattack.spawn.config.Configuration;
import eu.mhsl.craftattack.spawn.util.ComponentUtil;
import eu.mhsl.craftattack.spawn.util.IteratorUtil;
import net.kyori.adventure.sound.Sound;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import org.bukkit.*;
import org.bukkit.event.Listener;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.Map;

import static java.util.Map.entry;

public class Countdown extends Appliance {
    public Countdown() {
        super("countdown");
    }

    private int taskId = -1;
    private boolean isRunning = false;
    private final int countDownLength = localConfig().getInt("countdown");
    private int countDown;

    private int cycleState = 0;

    private World world = Bukkit.getWorld("world");
    private List<Material> glassColors = List.of(Material.RED_STAINED_GLASS, Material.YELLOW_STAINED_GLASS, Material.GREEN_STAINED_GLASS, Material.BLUE_STAINED_GLASS);
    private Material defaultGlassColor = Material.ORANGE_STAINED_GLASS;
    private Location glassLocation = new Location(world, -224, 67, -368);
    //
    private final Map<GameRule<Boolean>, Boolean> gamerulesAfterStart = Map.ofEntries(
            entry(GameRule.DO_DAYLIGHT_CYCLE, true),
            entry(GameRule.DO_INSOMNIA, true),
            entry(GameRule.ANNOUNCE_ADVANCEMENTS, true),
            entry(GameRule.DISABLE_RAIDS, false),
            entry(GameRule.DO_FIRE_TICK, true),
            entry(GameRule.DO_ENTITY_DROPS, true),
            entry(GameRule.DO_PATROL_SPAWNING, true),
            entry(GameRule.DO_TRADER_SPAWNING, true),
            entry(GameRule.DO_WEATHER_CYCLE, true),
            entry(GameRule.FALL_DAMAGE, true),
            entry(GameRule.FIRE_DAMAGE, true)
    );

    public void startCountdown() {
        if(!isEnabled()) return;
        if(isRunning) return;

        this.countDown = countDownLength;
        int startMusicAfter = 43;
        this.isRunning = true;
        this.taskId = Bukkit.getScheduler().scheduleSyncRepeatingTask(Main.instance(), () -> {
            if(this.countDown <= 0) {
                this.cancelCountdown();
                this.projectStart();
                return;
            }

            if(countDown == countDownLength - startMusicAfter) {
                world.playSound(glassLocation, org.bukkit.Sound.MUSIC_DISC_PIGSTEP, SoundCategory.RECORDS, 500f, 1f);
            }

            if(countDown > 60 && countDown % 60 == 0) {
                sendStartMessage(countDown / 60, "Minuten");
            }

            if(countDown <= 60 && (countDown <= 10 || countDown % 10 == 0)) {
                sendStartMessage(countDown, "Sekunden");
            }

            this.countDown--;
        }, 20, 20);
    }

    private void sendStartMessage(int timeLeft, String unit) {
        cycleBeacon();
        Bukkit.getOnlinePlayers().forEach(player -> player.sendMessage(
                Component.text()
                        .append(ComponentUtil.createRainbowText("CraftAttack", 10))
                        .append(Component.text(" startet in ", NamedTextColor.GOLD))
                        .append(Component.text(timeLeft, NamedTextColor.AQUA))
                        .append(Component.text(" " + unit + "!", NamedTextColor.GOLD))
        ));
    }

    private void cycleBeacon() {
        glassLocation.getBlock().setType(glassColors.get(cycleState));
        cycleState++;
        if(cycleState >= glassColors.size()) cycleState = 0;
    }

    public void cancelCountdown() {
        if(taskId == -1) return;
        Bukkit.getScheduler().cancelTask(this.taskId);
        this.isRunning = false;
        this.restoreBeforeStart();
    }

    public boolean isEnabled() {
        return localConfig().getBoolean("enabled");
    }

    public void setEnabled(boolean enabled) {
        localConfig().set("enabled", enabled);
        Configuration.saveChanges();
    }

    public void projectStart() {
        setEnabled(false);

        IteratorUtil.worlds(World::getWorldBorder, worldBorder -> worldBorder.setSize(worldBorder.getMaxSize()));
        IteratorUtil.worlds(world -> setGamerules(gamerulesAfterStart, false));

        Bukkit.getOnlinePlayers().forEach(player -> {
            player.setFoodLevel(20);
            player.setHealth(20);
            player.getInventory().clear();
            player.setGameMode(GameMode.SURVIVAL);
            player.setExp(0);
            player.setLevel(0);

            player.playSound(Sound.sound(org.bukkit.Sound.ITEM_GOAT_HORN_SOUND_5, Sound.Source.MASTER, 500f, 1f));

            player.sendMessage(Component.text("Viel Spaß bei CraftAttack!", NamedTextColor.GREEN));
        });

        Bukkit.getServer().advancementIterator().forEachRemaining(
                advancement -> Bukkit.getOnlinePlayers().forEach(
                        player -> player.getAdvancementProgress(advancement).getAwardedCriteria().forEach(
                                criteria -> player.getAdvancementProgress(advancement).revokeCriteria(criteria)
                        )
                )
        );

        glassLocation.getBlock().setType(defaultGlassColor);
    }

    public void restoreBeforeStart() {
        setEnabled(true);

        IteratorUtil.worlds(World::getWorldBorder, worldBorder -> {
            worldBorder.setSize(localConfig().getLong("worldborder-before"));
            worldBorder.setWarningDistance(0);
            worldBorder.setDamageAmount(0);
        });

        IteratorUtil.worlds(world -> world, world -> setGamerules(gamerulesAfterStart, true));

        glassLocation.getBlock().setType(defaultGlassColor);
    }

    private void setGamerules(Map<GameRule<Boolean>, Boolean> rules, boolean inverse) {
        rules.forEach((gameRule, value) -> IteratorUtil.worlds(world -> world.setGameRule(gameRule, value ^ inverse)));
    }

    public boolean isRunning() {
        return isRunning;
    }

    @Override
    protected @NotNull List<Listener> eventHandlers() {
        return List.of(new PlayerInvincibleListener());
    }

    @Override
    protected @NotNull List<ApplianceCommand<?>> commands() {
        return List.of(
                new ProjectStartCommand(),
                new ProjectStartCancelCommand(),
                new ProjectStartResetCommand()
        );
    }
}
